import 'dart:typed_data';
import 'package:vid/us/vid_us_data_http_reader.dart';
import 'package:vid/us/vid_us_data_reader.dart';
import 'package:vid/us/vid_us_image.dart';
import 'package:vid/us/vid_us_probe.dart';

enum VidUsImageFormat {
  Jpeg,
  Png,
  H264,
  Zip,
  Diff,
}

///Only suport read for now.
///Not suport carotid3d for now.
class VidUsImageData {
  late VidUsDataReader _reader;
  late List<int> _imagePositionList;
  late int _version;
  late int _imageCount;
  late VidUsProbe _probe;
  late VidUsImageFormat _imageFormat;
  late Uint8List _extendedData;

  final String header = "VINNO IMAGE DATA";

  /// Gets the version of this image data.
  int get version => _version;

  /// Gets the image count of this image data.
  int get imageCount => _imageCount;

  /// Gets the probe information.
  VidUsProbe get probe => _probe;

  /// Gets the image format of this image data.
  VidUsImageFormat get imageFormat => _imageFormat;

  /// Gets or sets the extended data.
  Uint8List get extendedData => _extendedData;

  /// Create a VINNO Image Data.
  VidUsImageData(Uint8List data) {
    _reader = VidUsDataReader(data);
    var header = _reader.readString();
    if (header != header) {
      throw Exception("The input data is not a VID data.");
    }
    _version = _reader.readInt();
    //Get probe info
    var probeData = _reader.readBytes();
    _probe = VidUsProbe.fromBytes(probeData);
    _imageFormat = VidUsImageFormat.values[_reader.readInt()];
    _extendedData = _reader.readBytes();

    _imagePositionList = [];
    var imagePositionListData = _reader.readBytes();
    _imageCount = imagePositionListData.length ~/ 8;
    var imagePositionReader = VidUsDataReader(imagePositionListData);
    for (var i = 0; i < _imageCount; i++) {
      _imagePositionList.add(imagePositionReader.readInt64V2());
    }
  }

  /// Get one image from the vid.
  VidUsImage getImage(int index) {
    if (index >= _imageCount || index < 0) {
      throw Exception("Can not find image Data");
    }
    //Jump to image.
    var imageData = _reader.readBytes(_imagePositionList[index]);
    return VidUsImage.fromBytes(imageData);
  }
}

class ReadTimeoutException implements Exception {}

class AlreadyClosedException implements Exception {}

///Only suport read for now.
///Not suport carotid3d for now.
class HttpVidUsImageData {
  late VidUsDataHttpReader _reader;
  late List<int> _imagePositionList;
  late int _version;
  late int _imageCount;
  late VidUsProbe _probe;
  late VidUsImageFormat _imageFormat;
  late Uint8List _extendedData;
  late int _readHeaderTimeout;
  late int _readImageTimeout;
  late bool _closed;

  final Duration delayDuration = const Duration(milliseconds: 10);
  final String header = "VINNO IMAGE DATA";

  /// Gets the version of this image data.
  int get version => _version;

  /// Gets the image count of this image data.
  int get imageCount => _imageCount;

  /// Gets the probe information.
  VidUsProbe get probe => _probe;

  /// Gets the image format of this image data.
  VidUsImageFormat get imageFormat => _imageFormat;

  /// Gets or sets the extended data.
  Uint8List get extendedData => _extendedData;

  /// Create a VINNO Image Data.
  /// [readHeaderTimeout] is the timeout value of reading header and probe information etc...
  /// [readImageTimeout] is the timeout value of reading one frame.
  HttpVidUsImageData(String url, {DownloadCallback? downloadCallback, int readHeaderTimeout = 1000, int readImageTimeout = 300}) {
    _closed = false;
    _readHeaderTimeout = readHeaderTimeout;
    _readImageTimeout = readImageTimeout;
    _reader = VidUsDataHttpReader(url, downloadCallback: downloadCallback);
  }

  //Initialize the StreamedVidUsImageData
  Future initialize() async {
    var header = await _readHeader();
    if (header != header) {
      throw Exception("The input data is not a VID data.");
    }
    _version = await _readVersion();
    //Get probe info
    _probe = await _readProbe();
    _imageFormat = await _readImageFormat();
    _extendedData = await _readExtendedData();
    _imagePositionList = [];
    var imagePositionListData = await _readImagePositionListData();
    _imageCount = imagePositionListData.length ~/ 8;
    var imagePositionReader = VidUsDataReader(imagePositionListData);
    for (var i = 0; i < _imageCount; i++) {
      _imagePositionList.add(imagePositionReader.readInt64V2());
    }
  }

  void close() {
    _closed = true;
    _reader.close();
  }

  //Read header from http data.
  Future<String> _readHeader() async {
    var timeout = 0;
    while (!_closed) {
      try {
        var header = _reader.readString();
        return header;
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  //Read the version from http data.
  Future<int> _readVersion() async {
    var timeout = 0;
    while (!_closed) {
      try {
        var version = _reader.readInt();
        return version;
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  //Read the probe info from http data.
  Future<VidUsProbe> _readProbe() async {
    var timeout = 0;
    while (!_closed) {
      try {
        var probeData = _reader.readBytes();
        return VidUsProbe.fromBytes(probeData);
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  //Read the image format from http data.
  Future<VidUsImageFormat> _readImageFormat() async {
    var timeout = 0;
    while (!_closed) {
      try {
        return VidUsImageFormat.values[_reader.readInt()];
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  //Read extended data from http data.
  Future<Uint8List> _readExtendedData() async {
    var timeout = 0;
    while (!_closed) {
      try {
        return _reader.readBytes();
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  //Read the image positions data from http data.
  Future<Uint8List> _readImagePositionListData() async {
    var timeout = 0;
    while (!_closed) {
      try {
        return _reader.readBytes();
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readHeaderTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
    throw AlreadyClosedException();
  }

  /// Get one image from the vid.
  Future<VidUsImage> getImage(int index) async {
    if (index >= _imageCount || index < 0) {
      throw Exception("Can not find image Data");
    }
    //Jump to image.
    var timeout = 0;
    while (!_closed) {
      try {
        var imageData = _reader.readBytes(_imagePositionList[index]);
        return VidUsImage.fromBytes(imageData);
      } catch (ex) {
        if (ex is NotReadyException) {
          if (timeout >= _readImageTimeout) {
            throw ReadTimeoutException();
          }
          await Future.delayed(delayDuration);
          timeout += delayDuration.inMilliseconds;
        } else {
          rethrow;
        }
      }
    }
      throw AlreadyClosedException();
  }
}
